;; Generated package description from embark-consult.el  -*- no-byte-compile: t -*-
(define-package "embark-consult" "0.5" "Consult integration for Embark" '((emacs "26.1") (embark "0.12") (consult "0.10")) :commit "5faf1389162dd64bfe3511dfb8f52c18efb5140b" :authors '(("Omar Antolín Camarena" . "omar@matem.unam.mx")) :maintainer '("Omar Antolín Camarena" . "omar@matem.unam.mx") :keywords '("convenience") :url "https://github.com/oantolin/embark")
